@@ -78,7 +78,12 @@ module Dnsruby
     attr_reader :port
     
     # Should TCP be used as a transport rather than UDP?
+    # If use_tcp==true, then ONLY TCP will be used as a transport.
     attr_reader :use_tcp
+
+    # If no_tcp==true, then ONLY UDP will be used as a transport.
+    # This should not generally be used, but is provided as a debugging aid.
+    attr_reader :no_tcp
     
     
     attr_reader :tsig
@@ -98,14 +103,17 @@ module Dnsruby
     
     # The current Config
     attr_reader :config
+
+    # Does this Resolver cache answers, and attempt to retrieve answer from the cache?
+    attr_reader :do_caching
     
     # The array of SingleResolvers used for sending query messages
     #    attr_accessor :single_resolvers # :nodoc:
     def single_resolvers=(s) # :nodoc:
       @configured = true
-#      @single_res_mutex.synchronize {
+      #      @single_res_mutex.synchronize {
       @single_resolvers = s
-#      }
+      #      }
     end
     def single_resolvers # :nodoc:
       if (!@configured)
@@ -139,6 +147,11 @@ module Dnsruby
     # requirements.
     attr_accessor :do_validation
     
+    # Defines whether we will cache responses, or pass every request to the
+    # upstream resolver.  This is only really useful when querying authoritative
+    # servers (as the upstream recursive resolver is likely to cache)
+    attr_accessor :do_caching
+
     #--
     #@TODO@ add load_balance? i.e. Target nameservers in a random, rather than pre-determined, order?
     #This is best done when configuring the Resolver, as it will re-order servers based on their response times.
@@ -156,6 +169,7 @@ module Dnsruby
     #   response = res.query("208.77.188.166", Types.PTR)
     def query(name, type=Types.A, klass=Classes.IN, set_cd=@dnssec)
       msg = Message.new
+      msg.do_caching = @do_caching
       msg.header.rd = 1
       msg.add_question(name, type, klass)
       msg.do_validation = @do_validation
@@ -167,6 +181,7 @@ module Dnsruby
 
     def query_no_validation_or_recursion(name, type=Types.A, klass=Classes.IN) # :nodoc: all
       msg = Message.new
+      msg.do_caching = @do_caching
       msg.header.rd = false
       msg.do_validation = false
       msg.add_question(name, type, klass)
@@ -278,7 +293,7 @@ module Dnsruby
     #* msg - the message to send
     #* client_queue - a Queue to push the response to, when it arrives
     #* client_query_id - an optional ID to identify the query to the client
-    #* use_tcp - whether to use TCP (defaults to SingleResolver.use_tcp)
+    #* use_tcp - whether to use only TCP (defaults to SingleResolver.use_tcp)
     #
     #Returns :
     # 
@@ -315,12 +330,19 @@ module Dnsruby
       if (!@configured)
         add_config_nameservers
       end
-#      @single_res_mutex.synchronize {
+      #      @single_res_mutex.synchronize {
       if (!@resolver_ruby) # @TODO@ Synchronize this?
         @resolver_ruby = ResolverRuby.new(self)
       end
-#      }
-      return @resolver_ruby.send_async(*args)
+      #      }
+      client_query_id = @resolver_ruby.send_async(*args)
+      if (@single_resolvers.length == 0)
+        Thread.start {
+          sleep(@query_timeout == 0 ? 1 : @query_timeout)
+          args[1].push([client_query_id, nil, ResolvTimeout.new("Query timed out - no nameservers configured")])
+        }
+      end
+      return client_query_id
     end
     
     # Close the Resolver. Unfinished queries are terminated with OtherResolvError.
@@ -347,12 +369,14 @@ module Dnsruby
     # * :query_timeout
     # * :retry_times
     # * :retry_delay
+    # * :do_caching
     def initialize(*args)
       # @TODO@ Should we allow :namesver to be an RRSet of NS records? Would then need to randomly order them?
       @resolver_ruby = nil
       @src_address = nil
       @single_res_mutex = Mutex.new
       @configured = false
+      @do_caching = true
       @config = Config.new()
       reset_attributes
       
@@ -396,14 +420,16 @@ module Dnsruby
       end
       @configured = true
       @single_res_mutex.synchronize {
-      # Add the Config nameservers
-      @config.nameserver.each do |ns|
-        @single_resolvers.push(PacketSender.new({:server=>ns, :dnssec=>@dnssec,
-              :use_tcp=>@use_tcp, :packet_timeout=>@packet_timeout,
-              :tsig => @tsig, :ignore_truncation=>@ignore_truncation,
-              :src_address=>@src_address, :src_port=>@src_port,
-              :recurse=>@recurse, :udp_size=>@udp_size}))
-      end
+        # Add the Config nameservers
+        @config.nameserver.each do |ns|
+          res = PacketSender.new({:server=>ns, :dnssec=>@dnssec,
+                :use_tcp=>@use_tcp, :no_tcp=>@no_tcp, :packet_timeout=>@packet_timeout,
+                :tsig => @tsig, :ignore_truncation=>@ignore_truncation,
+                :src_address=>@src_address, :src_port=>@src_port,
+                :do_caching=>@do_caching,
+                :recurse=>@recurse, :udp_size=>@udp_size})
+          @single_resolvers.push(res) if res
+        end
       }
     end
     
@@ -435,15 +461,17 @@ module Dnsruby
       @port = DefaultPort
       @udp_size = DefaultUDPSize
       @dnssec = DefaultDnssec
+      @do_caching= true
       @use_tcp = false
+      @no_tcp = false
       @tsig = nil
       @ignore_truncation = false
       @config = Config.new()
-      @src_address        = '0.0.0.0'
+      @src_address        = nil
       @src_port        = [0]
       @recurse = true
       @single_res_mutex.synchronize {
-      @single_resolvers=[]
+        @single_resolvers=[]
       }
       @configured = false
     end
@@ -451,6 +479,7 @@ module Dnsruby
     def update #:nodoc: all
       #Update any resolvers we have with the latest config
       @single_res_mutex.synchronize {
+        @single_resolvers.delete(nil) # Just in case...
         @single_resolvers.each do |res|
           update_internal_res(res)
         end
@@ -467,6 +496,7 @@ module Dnsruby
     def add_server(server)# :nodoc:
       @configured = true
       res = PacketSender.new(server)
+      raise ArgumentError.new("Can't create server #{server}") if !res
       update_internal_res(res)
       @single_res_mutex.synchronize {
         @single_resolvers.push(res)
@@ -474,7 +504,7 @@ module Dnsruby
     end
 
     def update_internal_res(res)
-      [:port, :use_tcp, :tsig, :ignore_truncation, :packet_timeout,
+      [:port, :use_tcp, :no_tcp, :tsig, :ignore_truncation, :packet_timeout,
         :src_address, :src_port, :recurse,
         :udp_size, :dnssec].each do |param|
 
@@ -487,9 +517,9 @@ module Dnsruby
     end
     def nameserver=(n)
       @configured = true
-            @single_res_mutex.synchronize {
-      @single_resolvers=[]
-            }
+      @single_res_mutex.synchronize {
+        @single_resolvers=[]
+      }
       set_config_nameserver(n)
       add_config_nameservers
     end
@@ -579,8 +609,10 @@ module Dnsruby
     end
 
     def Resolver.port_in_range(p)
-      if ((p == 0) || ((Iana::IANA_PORTS.index(p)) == nil &&
-              (p > 1024) && (p < 65535)))
+      if ((p == 0) || ((p >= 50000) && (p <= 65535)))
+        # @TODO@ IANA port bitmap - use 50000 - 65535 for now
+#            ((Iana::IANA_PORTS.index(p)) == nil &&
+#              (p > 1024) && (p < 65535)))
         return true
       end
       return false
@@ -602,12 +634,18 @@ module Dnsruby
       @use_tcp = on
       update
     end
+
+    def no_tcp=(on)
+      @no_tcp=on
+      update
+    end
     
     #Sets the TSIG to sign outgoing messages with.
     #Pass in either a Dnsruby::RR::TSIG, or a key_name and key (or just a key)
     #Pass in nil to stop tsig signing.
     #* res.tsig=(tsig_rr)
-    #* res.tsig=(key_name, key)
+    #* res.tsig=(key_name, key) # defaults to hmac-md5
+    #* res.tsig=(key_name, key, alg) # e.g. alg = "hmac-sha1"
     #* res.tsig=nil # Stop the resolver from signing
     def tsig=(t)
       @tsig=t
@@ -621,7 +659,11 @@ module Dnsruby
           if (args[0].instance_of?RR::TSIG)
             tsig = args[0]
           elsif (args[0].instance_of?Array)
-            tsig = RR.new_from_hash({:type => Types.TSIG, :klass => Classes.ANY, :name => args[0][0], :key => args[0][1]})
+            if (args[0].length > 2)
+              tsig = RR.new_from_hash({:type => Types.TSIG, :klass => Classes.ANY, :name => args[0][0], :key => args[0][1], :algorithm => args[0][2]})
+            else
+              tsig = RR.new_from_hash({:type => Types.TSIG, :klass => Classes.ANY, :name => args[0][0], :key => args[0][1]})
+            end
           end
         else
           #          Dnsruby.log.debug{"TSIG signing switched off"}
@@ -629,6 +671,8 @@ module Dnsruby
         end
       elsif (args.length ==2)
         tsig = RR.new_from_hash({:type => Types.TSIG, :klass => Classes.ANY, :name => args[0], :key => args[1]})
+      elsif (args.length ==3)
+        tsig = RR.new_from_hash({:type => Types.TSIG, :klass => Classes.ANY, :name => args[0], :key => args[1], :algorithm => args[2]})
       else
         raise ArgumentError.new("Wrong number of arguments to tsig=")
       end
@@ -661,6 +705,11 @@ module Dnsruby
       @persistent_udp = on
       update
     end
+
+    def do_caching=(on)
+      @do_caching=on
+      update
+    end
     
     def recurse=(a)
       @recurse = a
@@ -692,29 +741,28 @@ module Dnsruby
       #  e.g. timeouts[timeout1]=nameserver
       timeouts = {}
       retry_delay = @retry_delay
-#      @single_res_mutex.synchronize {
-        @retry_times.times do |retry_count|
-          if (retry_count>0)
-            retry_delay *= 2
-          end
-#          servers=[]
-#          @single_resolvers.each do |r| servers.push(r.server) end
-          @single_resolvers.each_index do |i|
-            res= @single_resolvers[i]
-            next if !res # @TODO@ WHY?1
-            offset = (i*@retry_delay.to_f/@single_resolvers.length)
-            if (retry_count==0)
-              timeouts[base+offset]=[res, retry_count]
-            else
-              if (timeouts.has_key?(base+retry_delay+offset))
-                Dnsruby.log.error{"Duplicate timeout key!"}
-                raise RuntimeError.new("Duplicate timeout key!")
-              end
-              timeouts[base+retry_delay+offset]=[res, retry_count]
+      #      @single_res_mutex.synchronize {
+      @retry_times.times do |retry_count|
+        if (retry_count>0)
+          retry_delay *= 2
+        end
+
+        @single_resolvers.delete(nil) # Just in case...
+        @single_resolvers.each_index do |i|
+          res= @single_resolvers[i]
+          offset = (i*@retry_delay.to_f/@single_resolvers.length)
+          if (retry_count==0)
+            timeouts[base+offset]=[res, retry_count]
+          else
+            if (timeouts.has_key?(base+retry_delay+offset))
+              Dnsruby.log.error{"Duplicate timeout key!"}
+              raise RuntimeError.new("Duplicate timeout key!")
             end
+            timeouts[base+retry_delay+offset]=[res, retry_count]
           end
         end
-#      }
+      end
+      #      }
       return timeouts      
     end
   end
@@ -729,7 +777,7 @@ module Dnsruby
     end
     def reset_attributes #:nodoc: all
       # data structures
-#      @mutex=Mutex.new
+      #      @mutex=Mutex.new
       @query_list = {}
       @timeouts = {}
     end
@@ -773,7 +821,7 @@ module Dnsruby
       
       tick_needed=false
       # add to our data structures
-#      @mutex.synchronize{
+      #      @mutex.synchronize{
       @parent.single_res_mutex.synchronize {
         tick_needed = true if @query_list.empty?
         if (@query_list.has_key?client_query_id)
@@ -813,7 +861,7 @@ module Dnsruby
     
     # Close the Resolver. Unfinished queries are terminated with OtherResolvError.
     def close
-#      @mutex.synchronize {
+      #      @mutex.synchronize {
       @parent.single_res_mutex.synchronize {
         @query_list.each do |client_query_id, values|
           msg, client_queue, q, outstanding = values
@@ -867,7 +915,7 @@ module Dnsruby
     def tick #:nodoc: all
       # Handle the tick
       # Do we have any retries due to be sent yet?
-#      @mutex.synchronize{
+      #      @mutex.synchronize{
       @parent.single_res_mutex.synchronize {
         time_now = Time.now
         @timeouts.keys.each do |client_query_id|
@@ -949,7 +997,7 @@ module Dnsruby
         Dnsruby.log.error{"Serious internal error!! #{id} expected, #{event_id} received"}
         raise RuntimeError.new("Serious internal error!! #{id} expected, #{event_id} received")
       end
-#      @mutex.synchronize{
+      #      @mutex.synchronize{
       @parent.single_res_mutex.synchronize {
         if (@query_list[client_query_id]==nil)
           #          print "Dead query response - ignoring\n"
@@ -1047,34 +1095,34 @@ module Dnsruby
     # TO BE CALLED IN A SYNCHRONIZED BLOCK
     def increment_resolver_priority(res)
       TheLog.debug("Incrementing resolver priority for #{res.server}\n")
-#      @parent.single_res_mutex.synchronize {
-        index = @parent.single_resolvers.index(res)
-        if (index > 0)
-          @parent.single_resolvers.delete(res)
-          @parent.single_resolvers.insert(index-1,res)
-        end
-#      }
+      #      @parent.single_res_mutex.synchronize {
+      index = @parent.single_resolvers.index(res)
+      if (index > 0)
+        @parent.single_resolvers.delete(res)
+        @parent.single_resolvers.insert(index-1,res)
+      end
+      #      }
     end
 
     # TO BE CALLED IN A SYNCHRONIZED BLOCK
     def decrement_resolver_priority(res)
       TheLog.debug("Decrementing resolver priority for #{res.server}\n")
-#      @parent.single_res_mutex.synchronize {
-        index = @parent.single_resolvers.index(res)
-        if (index < @parent.single_resolvers.length)
-          @parent.single_resolvers.delete(res)
-          @parent.single_resolvers.insert(index+1,res)
-        end
-#      }
+      #      @parent.single_res_mutex.synchronize {
+      index = @parent.single_resolvers.index(res)
+      if (index < @parent.single_resolvers.length)
+        @parent.single_resolvers.delete(res)
+        @parent.single_resolvers.insert(index+1,res)
+      end
+      #      }
     end
 
     # TO BE CALLED IN A SYNCHRONIZED BLOCK
     def demote_resolver(res)
       TheLog.debug("Demoting resolver priority for #{res.server} to bottom\n")
-#      @parent.single_res_mutex.synchronize {
-        @parent.single_resolvers.delete(res)
-        @parent.single_resolvers.push(res)
-#      }
+      #      @parent.single_res_mutex.synchronize {
+      @parent.single_resolvers.delete(res)
+      @parent.single_resolvers.push(res)
+      #      }
     end
     
     def handle_response(select_queue, query_id, response) #:nodoc: all
