@@ -99,7 +99,10 @@ module Dnsruby
           if (rr.type == Types::RRSIG)
             type_ok = (rr.type_covered == type)
           end
-          type_ok && (rr.klass == klass) && (rr.name.to_s.downcase == name.to_s.downcase)
+          if (!(/\.\z/ =~ name.to_s))
+            name = name.to_s + "."
+          end
+          type_ok && (rr.klass == klass) && (rr.name.to_s(true).downcase == name.to_s().downcase)
         }
         rrset = RRSet.new()
         rrs.each do |rr|
@@ -323,13 +326,15 @@ module Dnsruby
     
     # Return the rrsets of the specified type in the message
     def rrsets(type, klass=Classes::IN)
-      rrsets = []
+      rrsetss = []
       [@answer, @authority, @additional].each do |section|
-        if ((rrset = section.rrsets(type, klass)).length > 0)
-          rrsets.push(rrset)
+        if ((rrsets = section.rrsets(type, klass)).length > 0)
+          rrsets.each {|rrset|
+            rrsetss.push(rrset)
+          }
         end
       end
-      return rrsets
+      return rrsetss
     end
     
     # Return a hash, with the section as key, and the RRSets in that
@@ -666,10 +671,7 @@ module Dnsruby
     attr_accessor :cd
     
     #The Authenticated Data flag
-    attr_accessor :ad
-    
     #Relevant in DNSSEC context.
-    #
     #(The AD bit is only set on answers where signatures have been
     #cryptographically verified or the server is authoritative for the data
     #and is allowed to set the bit by policy.)
@@ -1061,8 +1063,10 @@ module Dnsruby
       put_length16 {rr.encode_rdata(self, canonical)}
     end
 
-    def put_name(d, canonical=false)
-      if (canonical)
+    def put_name(d, canonical=false, downcase = canonical)
+      # DNSSEC requires some records (e.g. NSEC, RRSIG) to be canonicalised, but
+      # not downcased. YUK!
+      if (downcase)
         d = d.downcase
       end
       put_labels(d.to_a, canonical)
