@@ -58,6 +58,8 @@ module Dnsruby
         @@queued_responses=[]
         @@queued_validation_responses=[]
         @@wakeup_sockets = get_socket_pair
+        # Suppress reverse lookups
+        BasicSocket.do_not_reverse_lookup = true
         #    end
         # Now start the select thread
         @@select_thread = Thread.new {
@@ -236,9 +238,16 @@ module Dnsruby
           answerip = msg.answerip.downcase
           answerfrom = msg.answerfrom.downcase
           dest_server = query_settings.dest_server
+          answeripaddr = IPAddr.new(answerip)
+          dest_server = IPAddr.new("0.0.0.0")
+          begin
+          destserveripaddr = IPAddr.new(dest_server)
+          rescue ArgumentError
+            # Host name not IP address
+          end
           if (dest_server && (dest_server != '0.0.0.0') &&
-                (answerip != query_settings.dest_server.downcase) &&
-                (answerfrom != query_settings.dest_server.downcase))
+                (answeripaddr != destserveripaddr) &&
+                (answerfrom != dest_server))
             Dnsruby.log.warn("Unsolicited response received from #{answerip} instead of #{query_settings.dest_server}")
           else 
             send_response_to_client(msg, bytes, socket)
@@ -276,7 +285,7 @@ module Dnsruby
           if (ret = res.check_response(msg, bytes, query, client_queue, id, tcp))
             remove_id(id)
             exception = msg.get_exception
-            if (ret.instance_of?TsigError)
+            if (ret.kind_of?TsigError)
               exception = ret
             end
             Dnsruby.log.debug{"Pushing response to client queue"}
@@ -401,6 +410,8 @@ module Dnsruby
             return
           end
         else
+          # @TODO@ Can we get recvfrom to stop issuing PTR queries when we already
+          # know both the FQDN and the IP address?
           if (ret = socket.recvfrom(packet_size))
             buf = ret[0]
             answerport=ret[1][1]
@@ -514,7 +525,11 @@ module Dnsruby
       # to send it out from its normal loop.
       Dnsruby.log.debug{"Pushing response to client queue direct from resolver or validator"}
       @@mutex.synchronize{
-        @@queued_responses.push([client_id, client_queue, msg, nil, query, res])
+        err = nil
+        if (msg.rcode == RCode.NXDOMAIN)
+          err = NXDomain.new
+        end
+        @@queued_responses.push([client_id, client_queue, msg, err, query, res])
       }
       # Make sure select loop is running!
       if (@@select_thread && @@select_thread.alive?)
